from google.cloud import storage
from google.cloud.storage import Client, Bucket

class Bucket(object):
    _client: Client = None
    _bucket: Bucket = None

    """
    This handles the bucket operations
    """
    def __init__(self, project_id, bucket_id):
        """
        Constructor for the Bucket class
        :param project_id: The project ID of the Google Cloud project
        :param bucket_id: The bucket ID of the Google Cloud Storage bucket
        :param retry: The retry object to use for the client
        """
        self._client = storage.Client(project=project_id)
        self._bucket = self._client.get_bucket(bucket_id)

    
    async def exists(self, path):
        """
        This method is used to check if the given path exists in the bucket
        :param path: The path to check if exists
        """
        blob = self._bucket.blob(path)
        return blob.exists()
    
    async def get(self, path):
        """
        This method is used to get the object from the bucket
        :param path: The path to get the object from
        """
        blob = self._bucket.blob(path)
        return blob.download_as_bytes()
    
    async def get_signed_url(self, path, method='GET', expiration=3600):
        """
        This method is used to generate a signed URL for given path and method
        :param path: The path for the object to generate the signed URL
        :param method: The method to generate the signed URL, default is GET
        :param expiration: The expiration time of the signed URL, default is 1 hour
        """
        blob = self._bucket.blob(path)
        return blob.generate_signed_url(method=method, expiration=expiration, version='v4')
    
    async def put(self, path, data, metadata=None):
        """
        This method is used to upload the object to the bucket
        :param path: The path to upload the object to
        :param data: The data to upload
        :param metadata: The metadata to upload with the object
        """
        blob = self._bucket.blob(path)
        blob.upload_from_string(data, content_type=metadata.get('Content-Type') if metadata else None)
    
    async def delete(self, path):
        """
        This method is used to delete the object from the bucket
        :param path: The path to delete the object from
        """
        blob = self._bucket.blob(path)
        blob.delete()




from json import loads, dumps
from os.path import splitext

from thumbor.storages import BaseStorage
from thumbor.utils import logger

from ..gcs.storage import GCSStorage

class Storage(GCSStorage, BaseStorage):
    """
    Google Cloud Storage (GCS) storage backend.
    """
    def __init__(self, context):
        """
        Constructor for the GcsStorage class.
        :param context: The Thumbor context to use for the storage backend.
        """
        BaseStorage.__init__(self, context)
        GCSStorage.__init__(self, context, 'TC_GCS_STORAGE')
        self.storage_expiration_seconds = context.config.get('STORAGE_EXPIRATION_SECONDS', 3600)

    async def put(self, path, file_bytes):
        """
        Put the given image data at the given path.
        :param path: The path to put the image data at.
        :param file_bytes: The image data to put at the given path.
        """
        try:
            await self._put(self._normalize_path(path), file_bytes)
        except Exception as e:
            logger.exception('Could not upload image %s to GCS: %s', path, e)
            return None
        
        return path

    async def put_crypto(self, path):
        """
        Stores crypto data at given path
        :param string path: Path to store the data at
        :return: Path where the crypto data is stored
        """
        if not self.context.config.STORES_CRYPTO_KEY_FOR_EACH_IMAGE:
            return

        if not self.context.server.security_key:
            raise RuntimeError(
                "STORES_CRYPTO_KEY_FOR_EACH_IMAGE can't be "
                "True if no SECURITY_KEY specified"
            )

        file_abspath = self._normalize_path(path)
        crypto_path = '%s.txt' % splitext(file_abspath)[0]

        try:
            await self._put(self.context.server.security_key.encode('utf-8'), crypto_path)
        except Exception as err:
            logger.exception('Unable to store crypto object: %s', err)
            return None

        logger.debug(
            "Stored crypto at %s (security key: %s)",
            crypto_path,
            self.context.server.security_key,
        )

        return file_abspath

    async def put_detector_data(self, path, data):
        """
        Stores detector data at given path
        :param string path: Path to store the data at
        :param string data:  Data to store
        :return: Path where the data is stored
        :rtype: string
        """
        file_abspath = self._normalize_path(path)

        path = '%s.detectors.txt' % splitext(file_abspath)[0]

        try:
            await self._put_object(dumps(data).encode('utf-8'), path)
        except Exception as err:
            logger.exception('Unable to store detector data: %s', err)
            return None

        return file_abspath

    async def get_crypto(self, path):
        """
        Retrieves crypto data at path
        :param string path: Path to search for crypto data
        """
        file_abspath = self._normalize_path(path)
        crypto_path = "%s.txt" % (splitext(file_abspath)[0])

        try:
            file_key = await self.storage.get(crypto_path)
        except Exception as err:
            logger.warn("[STORAGE] gcs blob not found at %s" % crypto_path)
            return None

        async with file_key['Body'] as stream:
            file_key = await stream.read()

        return file_key.decode('utf-8')

    async def get_detector_data(self, path):
        """
        Retrieves detector data from storage
        :param string path: Path where the data is stored
        """
        file_abspath = self._normalize_path(path)
        path = '%s.detectors.txt' % splitext(file_abspath)[0]

        try:
            file_key = await self.storage.get(path)
        except Exception:
            return None

        if not file_key or self.is_expired(file_key) or 'Body' not in file_key:
            return None

        async with file_key['Body'] as stream:
            return loads(await stream.read())

    async def get(self, path):
        """
        Gets data at path
        :param string path: Path for data
        """
        buffer = await super(Storage, self).get(path)
        return buffer


    async def exists(self, path):
        """
        Tells if data exists at given path
        :param string path: Path to check
        """
        file_abspath = self._normalize_path(path)
        return await self.storage.exists(file_abspath)

    async def remove(self, path):
        """
        Deletes data at path
        :param string path: Path to delete
        """
        return await self.storage.delete(self._normalize_path(path))



from hashlib import sha1
from os.path import join
from datetime import datetime
from dateutil.tz import tzutc

from .bucket import Bucket

class GCSStorage():
    """Google Cloud Storage (GCS) storage backend."""
    @property
    def is_auto_webp(self):
        """
        Determine if the storage backend should automatically convert images to WebP.
        :return: True if the storage backend should automatically convert images to WebP, False otherwise.
        """
        return self.context.config.AUTO_WEBP and hasattr(self.context, 'request') and self.context.request.accepts_webp
    
    @property
    def storage(self):
        """
        Get the storage object to use for the storage backend.
        :return: The storage object to use for the storage backend.
        """
        return Bucket(self._get_config('PROJECT_ID'), self._get_config('BUCKET_ID'))
    
    def __init__(self, context, config_prefix):
        """
        Constructor for the GcsStorage class.
        :param context: The Thumbor context to use for the storage backend.
        :param config_prefix: The prefix to use for the configuration keys.
        """
        self.context = context
        self.config_prefix = config_prefix

    async def get(self, path):
        """
        Get the image data for the given path.
        :param path: The path to get the image data for.
        :return: The image data for the given path.
        """
        return await self.storage.get(self._normalize_path(path))
    
    async def is_expired(self, path):
        """
        Tell if the image at the given path has expired.
        :param path: The path to check if the image has expired.
        :return: True if the image has expired, False otherwise.
        :rtype: bool
        """
        if path and 'LastModified' in path:
            expire_in_seconds = self.storage_expiration_seconds

            # Never expire
            if expire_in_seconds is None or expire_in_seconds == 0:
                return False

            timediff = datetime.now(tzutc()) - path['LastModified']

            return timediff.seconds > expire_in_seconds
        else:
            #If our key is bad just say we're expired
            return True
        
    async def _put(self, path, bytes, metadata=None):
        """
        Put the image data for the given path.
        :param path: The path to put the image data for.
        :param bytes: The image data to put.
        :param metadata: The metadata to put with the image data.
        """
        return await self.storage.put(path, bytes, metadata)
    
    def _get_config(self, config_key):
        """
        Get the configuration value for the given key.
        :param config_key: requested config
        :return: resolved config
        """
        return getattr(self.context.config, '{}_{}'.format(self.config_prefix, config_key))
    
    def _normalize_path(self, path):
        """
        Normalize the given path.
        :param path: The path to normalize.
        :return: The normalized path.
        """
        path = path.lstrip('/')  # Remove leading '/'
        path_segments = [path]

        root_path = self._get_config('ROOT_PATH')
        if root_path and root_path is not '':
            path_segments.insert(0, root_path)

        if self.is_auto_webp:
            path_segments.append("webp")

        if self._should_randomize_key():
            path_segments.insert(0, self._generate_digest(path_segments))

        normalized_path = join(path_segments[0], *path_segments[1:]).lstrip('/') if len(path_segments) > 1 else path_segments[0]
        if normalized_path.endswith('/'):
            normalized_path += self.context.config.TC_GCS_ROOT_IMAGE_NAME

        return normalized_path

    def _should_randomize_key(self):
        """
        Determine if the storage backend should randomize the image key.
        :return: True if the storage backend should randomize the image key, False otherwise.
        """
        return self.context.config.TC_GCS_RANDOMIZE_KEY
    
    def _generate_digest(self, segements):
        """
        Generate the digest for the given segments.
        :param segements: The segments to generate the digest for.
        :return: The digest for the given segments.
        """
        return sha1('.'.join(segements).encode('utf-8')).hexdigest()
    



__author__ = 'Ngo Sach Nhat (Chen)'

from thumbor.config import Config

Config.define('TC_GCS_STORAGE_PROJECT_ID', None, 'Google Cloud Storage project id', 'GCS')
Config.define('TC_GCS_STORAGE_BUCKET_ID', None, 'Google Cloud Storage bucket id', 'GCS')
Config.define('TC_GCS_STORAGE_ROOT_PATH', '', 'Google Cloud Storage path prefix', 'GCS')

Config.define('TC_GCS_LOADER_PROJECT_ID', None, 'Google Cloud Storage project id for loader', 'GCS')
Config.define('TC_GCS_LOADER_BUCKET_ID', None, 'Google Cloud Storage bucket id for loader', 'GCS')
Config.define('TC_GCS_LOADER_ROOT_PATH', '', 'Google Cloud Storage path prefix for loader', 'GCS')

Config.define('TC_GCS_RESULT_STORAGE_PROJECT_ID', None, 'Google Cloud Storage project id for result Storage', 'GCS')
Config.define('TC_GCS_RESULT_STORAGE_BUCKET_ID', None, 'Google Cloud Storage bucket id for result Storage', 'GCS')
Config.define('TC_GCS_RESULT_STORAGE_ROOT_PATH', '', 'Google Cloud Storage path prefix for result Storage', 'GCS')

Config.define('TC_GCS_ENABLE_HTTP_LOADER', False, 'Enable HTTP Loader as well?', 'GCS')
Config.define('TC_GCS_ALLOWED_BUCKETS', False, 'List of allowed buckets to be requested', 'GCS')
Config.define('TC_GCS_STORE_METADATA', False, 'Google Cloud Storage store result with metadata', 'GCS')
Config.define('TC_GCS_RANDOMIZE_KEYS', False, 'Should Google Cloud Storage keys be randomized? Defaults to False for BC, for performance, should be set to True', 'GCS')
Config.define('TC_GCS_ROOT_IMAGE_NAME', '', 'When resizing a URL that ends in a slash, what should the corresponding cache key be?', 'GCS')



from setuptools import setup, find_packages
import os

def version():
    """retrieve version from tag name"""
    github_ref_name = os.getenv('GITHUB_REF_NAME')
    if github_ref_name is not None:
        return github_ref_name
    return '0.0.1'

def readme():
    """print long description"""
    with open('README.md') as f:
        return f.read()

setup(
    name='tc_gcs',
    version=version(),
    packages=find_packages(),
    author='Ngo Sach Nhat (Chen)',
    author_email='nhatns.uet@gmail.com',
    description='A simple Thumbor custom storage for Google Cloud Storage',
    long_description=readme(),
    long_description_content_type='text/markdown',
    license='MIT',
    keywords='thumbor gcs google cloud storage',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    install_requires=[
        'thumbor>=7.6.0',
        'google-cloud-storage>=2.14.0',
    ],
)
